#pragma once

#include <atomic>
#include <map>
#include <set>
#include <string>
#include <vector>

#include <QAbstractListModel>
#include <QColor>
#include <QDateTime>
#include <QString>

#include <mtx/events/collections.hpp>

struct DescInfo {
    QString userid;
    QString body;
    QDateTime datetime;
};

struct MemberInfo {
    std::string avatar_url, display_name, userid;
};

class EventType : public QObject {
    Q_OBJECT

  public:
    EventType() : QObject() {}

    enum Type {
        // Unsupported event
        Unsupported,
        /// m.room_key_request
        KeyRequest,
        /// m.room.aliases
        Aliases,
        /// m.room.avatar
        Avatar,
        /// m.room.canonical_alias
        CanonicalAlias,
        /// m.room.create
        Create,
        /// m.room.encrypted.
        Encrypted,
        /// m.room.encryption.
        Encryption,
        /// m.room.guest_access
        GuestAccess,
        /// m.room.history_visibility
        HistoryVisibility,
        /// m.room.join_rules
        JoinRules,
        /// m.room.member
        Member,
        /// m.room.name
        Name,
        /// m.room.power_levels
        PowerLevels,
        /// m.room.tombstone
        Tombstone,
        /// m.room.topic
        Topic,
        /// m.room.redaction
        Redaction,
        /// m.room.pinned_events
        PinnedEvents,
        // m.sticker
        Sticker,
        // m.tag
        Tag,
        /// m.room.message
        AudioMessage,
        EmoteMessage,
        FileMessage,
        ImageMessage,
        LocationMessage,
        NoticeMessage,
        TextMessage,
        VideoMessage,
        UnknownMessage,
    };
    Q_ENUM(Type);
};

struct Room : public QAbstractListModel {
  private:
    Q_OBJECT
    Q_PROPERTY(QString name MEMBER name_ READ roomName WRITE setRoomName NOTIFY roomNameChanged)
    Q_PROPERTY(QString canonical_alias MEMBER canonical_alias_);
    Q_PROPERTY(QString replacement_room MEMBER replacement_room_);
    Q_PROPERTY(QString topic MEMBER topic_);
    Q_PROPERTY(QString avatar_url MEMBER avatar_url_ READ avatarUrl WRITE setAvatarUrl NOTIFY avatarChanged);
    Q_PROPERTY(bool is_invite MEMBER is_invite_ READ isInvite WRITE setIsInvite NOTIFY inviteStatusChanged);
    Q_PROPERTY(int member_count MEMBER member_count_);

  public:
    explicit Room(QObject *parent = 0);

    enum Roles {
        Type,
        Body,
        FormattedBody,
        UserId,
        UserName,
        Timestamp,
        Url,
        Height,
        Width,
        ProportionalHeight,
    };

    QHash<int, QByteArray> roleNames() const override;
    int rowCount(const QModelIndex &parent = QModelIndex()) const;
    QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const;

    Q_INVOKABLE QString userIdToUserName(QString id);
    Q_INVOKABLE QString roomName() const { return QString::fromStdString(name()); }
    Q_INVOKABLE QColor userColor(QString id, QColor background);

    std::string id;

    int unreadNotifications = 0;

    //! Who can access to the room.
    mtx::events::state::JoinRule join_rule = mtx::events::state::JoinRule::Public;
    bool guest_access = false;
    //! The list of tags associated with this room
    std::vector<std::string> tags;

    //! Metadata describing the last message in the timeline.
    DescInfo msgInfo;

    std::set<std::string> members;
    std::set<std::string> eventIds;
    std::map<std::string, MemberInfo> memberInfos;

    void applyEvents(const std::vector<mtx::events::collections::StateEvents> &state,
                     const std::vector<mtx::events::collections::TimelineEvents> &timeline, std::string prev_batch,
                     std::string next_batch) {
        emit newEvents(state, timeline, prev_batch, next_batch);
    }
    void applyTags(const std::vector<std::string> &tags) { emit setTags(tags); }
    void applyUnreadNotifications(int unread) { emit newUnreadState(unread); }
    std::string name() const;

  signals:
    void newEvents(std::vector<mtx::events::collections::StateEvents> state,
                   std::vector<mtx::events::collections::TimelineEvents> events, std::string prev_batch,
                   std::string next_batch);
    void newTags(std::vector<std::string> tags);
    void newUnreadState(int);
    void roomNameChanged(QString name);
    void avatarChanged(QString url);
    void inviteStatusChanged(bool isInvite);

    void markReadSig(int index);
    int lastReadSig();

  public slots:
    void addEvents(std::vector<mtx::events::collections::StateEvents> state,
                   std::vector<mtx::events::collections::TimelineEvents> timeline, std::string prev_batch,
                   std::string next_batch);
    void setTags(std::vector<std::string> tags) { this->tags = tags; }
    void setUnread(int unread) { this->unreadNotifications = unread; }
    void fetchOlderMessages();
    void sendTextMessage(QString message);
    void markRead(int index);
    int lastRead();

    void setAvatarUrl(QString url) {
        avatar_url_ = url;
        emit avatarChanged(avatar_url_);
    }
    QString avatarUrl() { return avatar_url_; }
    void setIsInvite(bool isInvite) {
        is_invite_ = isInvite;
        emit inviteStatusChanged(is_invite_);
    }
    bool isInvite() { return is_invite_; }
    void setRoomName(QString name) {
        name_ = name;
        emit roomNameChanged(name_);
    }

  private slots:
    void markReadSlot(int index);
    int lastReadSlot();

  private:
    std::vector<mtx::events::collections::TimelineEvents> events;
    std::string prev_batch;
    std::string next_batch;
    std::atomic<bool> loadingOlderMessages;

    std::map<QString, QColor> userColors;

    /// Properties
    //! The calculated name of the room.
    QString name_;
    QString canonical_alias_;
    QString replacement_room_;
    //! The topic of the room.
    QString topic_;
    //! The calculated avatar url of the room.
    QString avatar_url_;
    //! Whether or not the room is an invite.
    bool is_invite_ = false;
    //! Total number of members in the room.
    int member_count_ = 0;

    std::string lastRead_;
};
